UNIT LinEq; {Copyright 1992 Paul Milenkovic ALL RIGHTS RESERVED}

{Fast algorithm for solving linear systems of equations of the form

     R a + b = 0

 where "b" is a known row vector, "a" is the unknown row vector
 to be solved, and R is a symmetric positive definite matrix which is
 constructed from the list of generator vectors "g".}


INTERFACE USES DSPTypes;

CONST
    mrank  = 4;

TYPE
    g_col  = DSPTypes.coef_v0;
    g_list = ARRAY [1 .. mrank] OF g_col;


PROCEDURE FastSolve(VAR g: g_list; VAR b: g_col; ngp, ngn, n: INTEGER);


IMPLEMENTATION

CONST epsilon   =  1e-30;         {smallest allowed real number absolute value}


PROCEDURE FastSolve(VAR g: g_list; VAR b: g_col; ngp, ngn, n: INTEGER);

{The right hand side is placed in b[1] through b[n] and the solution
 appears in b[0] through b[n-1].

 For LPC analysis, place total signal energy in b[0]; minimum squared error
 will be returned in b[n].}

    VAR i, k: INTEGER;
        ssq: SINGLE;


    PROCEDURE CRotate(VAR g0, g1: g_col; i, n: INTEGER);

        VAR j: INTEGER;
            c, s, t, t1, t2: SINGLE;

    BEGIN
        IF Abs(g1[i]) > epsilon THEN BEGIN
            t := Sqrt(g0[i]*g0[i] + g1[i]*g1[i]);
            c := g0[i]/t;
            s := g1[i]/t;
            FOR j := 0 TO n DO BEGIN
                t2    := c*g1[j] - s*g0[j];
                g0[j] := c*g0[j] + s*g1[j];
                g1[j] := t2;
            END;
        END;
    END;


    PROCEDURE HRotate(VAR g0, g1: g_col; i, n: INTEGER);

        VAR j: INTEGER;
            c, s, t, t1, t2: DOUBLE;

    BEGIN
        t := Sqrt(g0[i]*g0[i] - g1[i]*g1[i]);
        c := g0[i]/t;
        s := g1[i]/t;
        FOR j := 0 TO n DO BEGIN
            t2    := c*g1[j] - s*g0[j];
            g0[j] := c*g0[j] - s*g1[j];
            g1[j] := t2;
        END;
    END;


    PROCEDURE BackSubstitute(VAR g, b: g_col; VAR ssq: SINGLE; i, n: INTEGER);

        VAR j: INTEGER;
            s: SINGLE;

    BEGIN
        s := b[i]/g[i];
        ssq := ssq - s*s;
        FOR j := 0 TO n DO b[j] := b[j] - s*g[j];
        FOR j := n DOWNTO 1 DO g[j] := g[j-1];
        g[0] := 0;
    END;

BEGIN
    ssq := b[0];
    b[0] := 0;
    FOR i := 1 TO n DO BEGIN
        FOR k := 2 TO ngp DO CRotate(g[1],g[k],i,n);
        FOR k := ngp+2 TO ngp+ngn DO CRotate(g[ngp+1],g[k],i,n);
        IF ngn > 0 THEN HRotate(g[1],g[ngp+1],i,n);
        BackSubstitute(g[1],b,ssq,i,n)
    END;
    b[n] := ssq;
END;

END.
